
// Suppose we have two functions that both update a shared
// resource and we cannot allow a race condition.
atomicUpdateA(Data data)
{
   pthread_mutex_lock(&mutex);
   // enter critical section,
   // do update
   pthread_mutex_unlock(&mutex);
}

atomicUpdateB(Data data)
{
   pthread_mutex_lock(&mutex);
   // enter critical section,
   // do update
   pthread_mutex_unlock(&mutex);
}


// Now suppose that at some later date it is
// decided that atomicUpdateA() must also call
// atomicUpdateB() in the middle of its work.
// We could do this
atomicUpdateA(Data data)
{
   pthread_mutex_lock(&mutex);
   // enter critical section,
   // do some updating
   pthread_mutex_unlock(&mutex);
   atomicUpdateB(data);
   pthread_mutex_lock(&mutex);
   // enter critical section,
   // do some more updating
   pthread_mutex_unlock(&mutex);
}
// but then the work that atomicUpdateA does is not "atomic".
// Another thread can interrupt atomicUpdateA in the middle
// of its work.

// So we must have the following.

atomicUpdateA(Data data)
{
   pthread_mutex_lock(&mutex);
   // enter critical section,
   // do some updating
   atomicUpdateB(data);
   // do some more updating
   pthread_mutex_unlock(&mutex);
}

atomicUpdateB(Data data)
{
   pthread_mutex_lock(&mutex);
   // enter critical section,
   // do update
   pthread_mutex_unlock(&mutex);
}

// Recall that pthreads has four kinds of mutex
//   http://pubs.opengroup.org/onlinepubs/9699919799/functions/pthread_mutex_lock.html
// They are
//    default
//    normal
//    error-check
//    recursive
// How does the above code work with each kind of mutex?
//
// With default mutex, the critical section is violated
// when atomicUpdateB() returns to atomicUpdateA().
//
// With normal mutex, the code deadlocks as soon as
// atomicUpdateA() calls atomicUpdateB() and it tires
// relock the lock.
//
// With errorcheck mutex, atomicUpdateB() would need to
// be rewritten (which may not be possible) so that it
// does not unlock the lock if it finds out that it
// already owned the lock. Otherwise, this mutex will
// work like the default mutex.
//
// With the recursive mutex, the code above will work
// fine, and this is exactly the case that recursive
// mutexes were designed for.

// NOTE: We cannot drop the mutex lock from atomicUpdateB()
// because there may be threads in our code that call it
// independently of atomicUpdateA().
